'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var event = require('@hmans/event');

var isArchetype = function isArchetype(entity) {
  for (var _len = arguments.length, properties = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    properties[_key - 1] = arguments[_key];
  }

  for (var _i = 0, _properties = properties; _i < _properties.length; _i++) {
    var key = _properties[_i];
    if (!(key in entity)) return false;
  }

  return true;
};
var archetypeCache = new Map();
var archetype = function archetype() {
  for (var _len2 = arguments.length, properties = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    properties[_key2] = arguments[_key2];
  }

  var normalizedProperties = properties.sort().filter(function (p) {
    return !!p && p !== "";
  });
  var key = JSON.stringify(normalizedProperties);
  if (archetypeCache.has(key)) return archetypeCache.get(key);

  var predicate = function predicate(entity) {
    return isArchetype.apply(void 0, [entity].concat(properties));
  };

  archetypeCache.set(key, predicate);
  return predicate;
};

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _createForOfIteratorHelper(o, allowArrayLike) {
  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];

  if (!it) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;

      var F = function () {};

      return {
        s: F,
        n: function () {
          if (i >= o.length) return {
            done: true
          };
          return {
            done: false,
            value: o[i++]
          };
        },
        e: function (e) {
          throw e;
        },
        f: F
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var normalCompletion = true,
      didErr = false,
      err;
  return {
    s: function () {
      it = it.call(o);
    },
    n: function () {
      var step = it.next();
      normalCompletion = step.done;
      return step;
    },
    e: function (e) {
      didErr = true;
      err = e;
    },
    f: function () {
      try {
        if (!normalCompletion && it.return != null) it.return();
      } finally {
        if (didErr) throw err;
      }
    }
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  Object.defineProperty(Constructor, "prototype", {
    writable: false
  });
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

var _Symbol$iterator;
_Symbol$iterator = Symbol.iterator;
var Bucket = /*#__PURE__*/function () {
  function Bucket() {
    _classCallCheck(this, Bucket);

    _defineProperty(this, "entities", new Array());

    _defineProperty(this, "onEntityAdded", new event.Event());

    _defineProperty(this, "onEntityRemoved", new event.Event());

    _defineProperty(this, "onEntityTouched", new event.Event());

    _defineProperty(this, "derivedBuckets", new WeakMap());
  }

  _createClass(Bucket, [{
    key: _Symbol$iterator,
    value: function value() {
      return this.entities[Symbol.iterator]();
    }
  }, {
    key: "has",
    value: function has(entity) {
      return this.entities.includes(entity);
    }
  }, {
    key: "add",
    value: function add(entity) {
      var index = this.entities.indexOf(entity);
      /* Add the entity if we don't already have it */

      if (index === -1) {
        this.entities.push(entity);
        this.onEntityAdded.emit(entity);
      }

      return entity;
    }
  }, {
    key: "touch",
    value: function touch(entity) {
      var index = this.entities.indexOf(entity);

      if (index !== -1) {
        /* Emit an event if the entity changed */
        this.onEntityTouched.emit(entity);
      }

      return entity;
    }
  }, {
    key: "remove",
    value: function remove(entity) {
      /* Only act if we know about the entity */
      var index = this.entities.indexOf(entity);
      if (index === -1) return;
      /* Remove entity from our list */

      this.entities[index] = this.entities[this.entities.length - 1];
      this.entities.pop();
      /* Emit event */

      this.onEntityRemoved.emit(entity);
    }
  }, {
    key: "clear",
    value: function clear() {
      for (var i = this.entities.length - 1; i >= 0; i--) {
        this.remove(this.entities[i]);
      }
    }
  }, {
    key: "derive",
    value: function derive() {
      var predicate = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : function () {
        return true;
      };

      /* Check if we already have a derived bucket for this predicate */
      if (this.derivedBuckets.has(predicate)) {
        return this.derivedBuckets.get(predicate);
      }
      /* Create bucket */


      var bucket = new Bucket();
      /* Add to cache */

      this.derivedBuckets.set(predicate, bucket);
      /* Add entities that match the predicate */

      var _iterator = _createForOfIteratorHelper(this.entities),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          var _entity = _step.value;
          if (predicate(_entity)) bucket.add(_entity);
        }
        /* Listen for new entities */

      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }

      this.onEntityAdded.addListener(function (entity) {
        if (predicate(entity)) bucket.add(entity);
      });
      /* Listen for removed entities */

      this.onEntityRemoved.addListener(function (entity) {
        bucket.remove(entity);
      });
      /* Listen for changed entities */

      this.onEntityTouched.addListener(function (entity) {
        if (predicate(entity)) bucket.add(entity) && bucket.touch(entity);else bucket.remove(entity);
      });
      return bucket;
    }
  }]);

  return Bucket;
}();

var entityToId = new WeakMap();
var nextId = 0;
var id = function id(entity) {
  if (!entityToId.has(entity)) entityToId.set(entity, nextId++);
  return entityToId.get(entity);
};

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };
  return _setPrototypeOf(o, p);
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  Object.defineProperty(subClass, "prototype", {
    writable: false
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf.bind() : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  } else if (call !== void 0) {
    throw new TypeError("Derived constructors may only return object or undefined");
  }

  return _assertThisInitialized(self);
}

function _createSuper(Derived) {
  var hasNativeReflectConstruct = _isNativeReflectConstruct();
  return function _createSuperInternal() {
    var Super = _getPrototypeOf(Derived),
        result;

    if (hasNativeReflectConstruct) {
      var NewTarget = _getPrototypeOf(this).constructor;
      result = Reflect.construct(Super, arguments, NewTarget);
    } else {
      result = Super.apply(this, arguments);
    }

    return _possibleConstructorReturn(this, result);
  };
}

var World = /*#__PURE__*/function (_Bucket) {
  _inherits(World, _Bucket);

  var _super = _createSuper(World);

  function World() {
    _classCallCheck(this, World);

    return _super.apply(this, arguments);
  }

  _createClass(World, [{
    key: "addProperty",
    value: function addProperty(entity, property, value) {
      if (property in entity) return;
      entity[property] = value;
      this.touch(entity);
    }
  }, {
    key: "removeProperty",
    value: function removeProperty(entity, property) {
      if (!(property in entity)) return;
      delete entity[property];
      this.touch(entity);
    }
  }, {
    key: "archetype",
    value: function archetype$1() {
      return this.derive(archetype.apply(void 0, arguments));
    }
  }]);

  return World;
}(Bucket);

exports.Bucket = Bucket;
exports.World = World;
exports.archetype = archetype;
exports.id = id;
exports.isArchetype = isArchetype;
